#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <dirent.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>

#include "common.h"
#include "macros.h"
#include "failures.h"
#include "xmalloc.h"
#include "vm.h"
#include "vmlibcall.h"

#define VM_LIBCALL_INIT	"vm_libcall_init"

#define	TABLE_SIZE	0xff

static struct vm_libcall *vm_libcall_table[TABLE_SIZE];

/* Load all dynamic libraries, locate `syscall_init' symbol address,
 * and call initialization function. */
ret_t vm_libcall_load(struct vm *machine, const char *path_to_lib)
{
	struct dirent *entry;
	DIR *dir;
	void (*init)();
	char *error;
	char *path;
	int ret;

	return_val_if_fail(machine != NULL, ret_err);
	return_val_if_fail(path_to_lib != NULL, ret_err);

	dir = opendir(path_to_lib);
	if (dir == NULL)
		xerror(1, "can't open library calls directory %s: %s",
				path_to_lib, strerror(errno));
	
	while ((entry = readdir(dir)) != NULL) {
		struct stat st;
		size_t len;
		char *ch;
		void *handle;

		len = strlen(entry->d_name);
		if (len < 4)
			continue;

		ch = &entry->d_name[len - 3];
		if (ch[0] != '.' || ch[1] != 's' || ch[2] != 'o')
			continue;

		asprintf(&path, "%s/%s", path_to_lib, entry->d_name);
		ret = lstat(path, &st);
		if (ret != 0 || !S_ISREG(st.st_mode)) {
			xfree(path);
			continue;
		}

		if (machine->dl_used == machine->dl_size) {
			machine->dl_handle = xrealloc0(machine->dl_handle,
					machine->dl_size * sizeof(void *),
					(8 + machine->dl_size) * sizeof(void *));
			machine->dl_size += 8;
		}

		/* Resolve all symbols, we can't use lazy symbol resolver */
		handle = dlopen(path, RTLD_NOW);
		if (handle == NULL) {
			xwarning("%s", dlerror());
			xfree(path);
			continue;
		}

		machine->dl_handle[machine->dl_used++] = handle;
		xfree(path);
		/* Clear errors. */
		dlerror();
		init = dlsym(handle, VM_LIBCALL_INIT);

		error = dlerror();
		if (error != NULL) {
			xwarning("%s", error);
			continue;
		}
		/* Call vm_libcall_init() routine. */
		init();
	}
	
	closedir(dir);
	if (machine->dl_used == 0)
		xwarning("can't load any libraries");

	return ret_ok;
}

ret_t vm_libcall_unload(struct vm *machine)
{
	int res;
	unsigned int i;

	return_val_if_fail(machine != NULL, ret_err);
	return_val_if_fail(machine->dl_handle != NULL, ret_err);

	for (i = 0; i < machine->dl_used; i++) {
		res = dlclose(machine->dl_handle[i]);
		if (res != 0)
			return ret_err;
	}

	if (machine->dl_handle != NULL) {
		xfree(machine->dl_handle);
		machine->dl_handle = NULL;
	}

	return ret_ok;
}

ret_t vm_libcall_register(struct vm_libcall *q)
{
	return_val_if_fail(q->name != NULL, ret_err);
	return_val_if_fail(q->handler != NULL, ret_err);
	
	if (q->num >= TABLE_SIZE)
		return ret_err;
	
	if (vm_libcall_table[q->num] != NULL) {
		xwarning("system call %d is already initialized", q->num);
		return ret_err;
	}

	vm_libcall_table[q->num] = q;

	return ret_ok;			
}

ret_t vm_libcall_call_by_num(struct vm *machine, u_int32_t num)
{
	return_val_if_fail(machine != NULL, ret_err);

	if (num >= TABLE_SIZE)
		return ret_err;

	if (vm_libcall_table[num] == NULL)
		return ret_err;

	vm_libcall_table[num]->handler(machine);

	return ret_ok;
}

